// Copyright (c) Microsoft Corporation.
// Licensed under the MIT License.
import path from "node:path";
import { runGenerator } from "./generator.js";
import { runTemplates } from "./template.js";
import log from "../utils/log.js";
var SuggestionIcons;
(function (SuggestionIcons) {
    SuggestionIcons["File"] = "\uD83D\uDCC4";
    SuggestionIcons["Folder"] = "\uD83D\uDCC1";
    SuggestionIcons["Subcommand"] = "\uD83D\uDCE6";
    SuggestionIcons["Option"] = "\uD83D\uDD17";
    SuggestionIcons["Argument"] = "\uD83D\uDCB2";
    SuggestionIcons["Mixin"] = "\uD83C\uDFDD\uFE0F";
    SuggestionIcons["Shortcut"] = "\uD83D\uDD25";
    SuggestionIcons["Special"] = "\u2B50";
    SuggestionIcons["Default"] = "\uD83D\uDCC0";
})(SuggestionIcons || (SuggestionIcons = {}));
const getIcon = (icon, suggestionType) => {
    // TODO: enable fig icons once spacing is better
    // if (icon && /[^\u0000-\u00ff]/.test(icon)) {
    //   return icon;
    // }
    switch (suggestionType) {
        case "arg":
            return SuggestionIcons.Argument;
        case "file":
            return SuggestionIcons.File;
        case "folder":
            return SuggestionIcons.Folder;
        case "option":
            return SuggestionIcons.Option;
        case "subcommand":
            return SuggestionIcons.Subcommand;
        case "mixin":
            return SuggestionIcons.Mixin;
        case "shortcut":
            return SuggestionIcons.Shortcut;
        case "special":
            return SuggestionIcons.Special;
    }
    return SuggestionIcons.Default;
};
const getLong = (suggestion) => {
    return suggestion instanceof Array ? suggestion.reduce((p, c) => (p.length > c.length ? p : c)) : suggestion;
};
const getPathy = (type) => {
    return type === "file" || type === "folder";
};
const toSuggestion = (suggestion, name, type) => {
    if (suggestion.name == null)
        return;
    return {
        name: name ?? getLong(suggestion.name),
        description: suggestion.description,
        icon: getIcon(suggestion.icon, type ?? suggestion.type),
        allNames: suggestion.name instanceof Array ? suggestion.name : [suggestion.name],
        priority: suggestion.priority ?? 50,
        insertValue: suggestion.insertValue,
        pathy: getPathy(suggestion.type),
    };
};
function filter(suggestions, filterStrategy, partialCmd, suggestionType) {
    if (!partialCmd)
        return suggestions.map((s) => toSuggestion(s, undefined, suggestionType)).filter((s) => s != null);
    switch (filterStrategy) {
        case "fuzzy":
            return suggestions
                .map((s) => {
                if (s.name == null)
                    return;
                if (s.name instanceof Array) {
                    const matchedName = s.name.find((n) => n.toLowerCase().includes(partialCmd.toLowerCase()));
                    return matchedName != null
                        ? {
                            name: matchedName,
                            description: s.description,
                            icon: getIcon(s.icon, s.type ?? suggestionType),
                            allNames: s.name,
                            priority: s.priority ?? 50,
                            insertValue: s.insertValue,
                            pathy: getPathy(s.type),
                        }
                        : undefined;
                }
                return s.name.toLowerCase().includes(partialCmd.toLowerCase())
                    ? {
                        name: s.name,
                        description: s.description,
                        icon: getIcon(s.icon, s.type ?? suggestionType),
                        allNames: [s.name],
                        priority: s.priority ?? 50,
                        insertValue: s.insertValue,
                        pathy: getPathy(s.type),
                    }
                    : undefined;
            })
                .filter((s) => s != null);
        default:
            return suggestions
                .map((s) => {
                if (s.name == null)
                    return;
                if (s.name instanceof Array) {
                    const matchedName = s.name.find((n) => n.toLowerCase().startsWith(partialCmd.toLowerCase()));
                    return matchedName != null
                        ? {
                            name: matchedName,
                            description: s.description,
                            icon: getIcon(s.icon, s.type ?? suggestionType),
                            allNames: s.name,
                            insertValue: s.insertValue,
                            priority: s.priority ?? 50,
                            pathy: getPathy(s.type),
                        }
                        : undefined;
                }
                return s.name.toLowerCase().startsWith(partialCmd.toLowerCase())
                    ? {
                        name: s.name,
                        description: s.description,
                        icon: getIcon(s.icon, s.type ?? suggestionType),
                        allNames: [s.name],
                        insertValue: s.insertValue,
                        priority: s.priority ?? 50,
                        pathy: getPathy(s.type),
                    }
                    : undefined;
            })
                .filter((s) => s != null);
    }
}
const generatorSuggestions = async (generator, acceptedTokens, filterStrategy, partialCmd, cwd) => {
    const generators = generator instanceof Array ? generator : generator ? [generator] : [];
    const tokens = acceptedTokens.map((t) => t.token);
    if (partialCmd)
        tokens.push(partialCmd);
    const suggestions = (await Promise.all(generators.map((gen) => runGenerator(gen, tokens, cwd)))).flat();
    return filter(suggestions.map((suggestion) => ({ ...suggestion, priority: suggestion.priority ?? 60 })), filterStrategy, partialCmd, undefined);
};
const templateSuggestions = async (templates, filterStrategy, partialCmd, cwd) => {
    return filter(await runTemplates(templates ?? [], cwd), filterStrategy, partialCmd, undefined);
};
const suggestionSuggestions = (suggestions, filterStrategy, partialCmd) => {
    const cleanedSuggestions = suggestions?.map((s) => (typeof s === "string" ? { name: s } : s)) ?? [];
    return filter(cleanedSuggestions ?? [], filterStrategy, partialCmd, undefined);
};
const subcommandSuggestions = (subcommands, filterStrategy, partialCmd) => {
    return filter(subcommands ?? [], filterStrategy, partialCmd, "subcommand");
};
const optionSuggestions = (options, acceptedTokens, filterStrategy, partialCmd) => {
    const usedOptions = new Set(acceptedTokens.filter((t) => t.isOption).map((t) => t.token));
    const validOptions = options?.filter((o) => o.exclusiveOn?.every((exclusiveOption) => !usedOptions.has(exclusiveOption)) ?? true);
    return filter(validOptions ?? [], filterStrategy, partialCmd, "option");
};
const getEscapedPath = (value) => {
    return value?.replaceAll(" ", "\\ ");
};
function adjustPathSuggestions(suggestions, partialToken) {
    if (partialToken == null || partialToken.isQuoted)
        return suggestions;
    return suggestions.map((s) => s.pathy ? { ...s, insertValue: getEscapedPath(s.insertValue), name: s.insertValue == null ? getEscapedPath(s.name) : s.name } : s);
}
const removeAcceptedSuggestions = (suggestions, acceptedTokens) => {
    const seen = new Set(acceptedTokens.map((t) => t.token));
    return suggestions.filter((s) => s.allNames.every((n) => !seen.has(n)));
};
const removeDuplicateSuggestion = (suggestions) => {
    const seen = new Set();
    return suggestions
        .map((s) => {
        if (seen.has(s.name))
            return null;
        seen.add(s.name);
        return s;
    })
        .filter((s) => s != null);
};
const removeEmptySuggestion = (suggestions) => {
    return suggestions.filter((s) => s.name.length > 0);
};
export const getSubcommandDrivenRecommendation = async (subcommand, persistentOptions, partialToken, argsDepleted, argsFromSubcommand, acceptedTokens, cwd) => {
    log.debug({ msg: "suggestion point", subcommand, persistentOptions, partialToken, argsDepleted, argsFromSubcommand, acceptedTokens, cwd });
    if (argsDepleted && argsFromSubcommand) {
        return;
    }
    let partialCmd = partialToken?.token;
    if (partialToken?.isPath) {
        partialCmd = partialToken.isPathComplete ? "" : path.basename(partialCmd ?? "");
    }
    const suggestions = [];
    const argLength = subcommand.args instanceof Array ? subcommand.args.length : subcommand.args ? 1 : 0;
    const allOptions = persistentOptions.concat(subcommand.options ?? []);
    if (!argsFromSubcommand) {
        suggestions.push(...subcommandSuggestions(subcommand.subcommands, subcommand.filterStrategy, partialCmd));
        suggestions.push(...optionSuggestions(allOptions, acceptedTokens, subcommand.filterStrategy, partialCmd));
    }
    if (argLength != 0) {
        const activeArg = subcommand.args instanceof Array ? subcommand.args[0] : subcommand.args;
        suggestions.push(...(await generatorSuggestions(activeArg?.generators, acceptedTokens, activeArg?.filterStrategy, partialCmd, cwd)));
        suggestions.push(...suggestionSuggestions(activeArg?.suggestions, activeArg?.filterStrategy, partialCmd));
        suggestions.push(...(await templateSuggestions(activeArg?.template, activeArg?.filterStrategy, partialCmd, cwd)));
    }
    return {
        suggestions: removeDuplicateSuggestion(removeEmptySuggestion(removeAcceptedSuggestions(adjustPathSuggestions(suggestions.sort((a, b) => b.priority - a.priority), partialToken), acceptedTokens))),
    };
};
export const getArgDrivenRecommendation = async (args, subcommand, persistentOptions, partialToken, acceptedTokens, variadicArgBound, cwd) => {
    let partialCmd = partialToken?.token;
    if (partialToken?.isPath) {
        partialCmd = partialToken.isPathComplete ? "" : path.basename(partialCmd ?? "");
    }
    const activeArg = args[0];
    const allOptions = persistentOptions.concat(subcommand.options ?? []);
    const suggestions = [
        ...(await generatorSuggestions(args[0].generators, acceptedTokens, activeArg?.filterStrategy, partialCmd, cwd)),
        ...suggestionSuggestions(args[0].suggestions, activeArg?.filterStrategy, partialCmd),
        ...(await templateSuggestions(args[0].template, activeArg?.filterStrategy, partialCmd, cwd)),
    ];
    if (activeArg.isOptional || (activeArg.isVariadic && variadicArgBound)) {
        suggestions.push(...subcommandSuggestions(subcommand.subcommands, activeArg?.filterStrategy, partialCmd));
        suggestions.push(...optionSuggestions(allOptions, acceptedTokens, activeArg?.filterStrategy, partialCmd));
    }
    return {
        suggestions: removeDuplicateSuggestion(removeEmptySuggestion(removeAcceptedSuggestions(adjustPathSuggestions(suggestions.sort((a, b) => b.priority - a.priority), partialToken), acceptedTokens))),
        argumentDescription: activeArg.description ?? activeArg.name,
    };
};
